using Avalonia;
using Avalonia.Controls;
using Avalonia.Controls.Shapes;
using Avalonia.Markup.Xaml;
using Avalonia.Media;
using Avalonia.Threading;
using System;
using System.Timers;

namespace TDP.Clock
{
    public partial class MainWindow : Window
    {
        // Define some constants needed to draw the clock and more
        private const double _flatCornerDegrees = 180;

        private const int _clockMinutes = 60;
        private const double _clockBorderTickness = 3;
        private const double _hourTickness = 15;
        private const double _minuteTickness = 3;

        private const double _handHoursTickness = 25;
        private const double _handMinutesTickness = 10;
        private const double _handSecondsTickness = 10;

        private const double _degreesPerSecond = 6;
        private const double _degreesPerMinute = 6;
        private const double _degreesPerHour = 30;
        private const int _clockFreqUpdatesSeconds = 1;

        private const double _secondsPerMinute = 60;
        private const double _minutesPerHour = 60;

        private Canvas _canvas;
        private Ellipse _clockBorder;
        private Line[] _clockMarkers;
        private Line _handHours;
        private Line _handMinutes;
        private Line _handSeconds;

        // This will be used to trigger clock refresh
        private DispatcherTimer _timer;

        public MainWindow()
        {
            InitializeComponent();
#if DEBUG
            this.AttachDevTools();
#endif
        }

        private double GetClockRadius()
        {
            return (ClientSize.Width > ClientSize.Height ? ClientSize.Height : ClientSize.Width) / 2;
        }

        private double GetClockSize()
        {
            return ClientSize.Width > ClientSize.Height ? ClientSize.Height : ClientSize.Width;
        }

        private Point GetClockCenteredPosition()
        {
            return new Point((ClientSize.Width - GetClockSize()) / 2, (ClientSize.Height - GetClockSize()) / 2);
        }

        private Point GetNewOrigin()
        {
            return new Point(ClientSize.Width / 2, ClientSize.Height / 2);
        }

        
        private Point TranslatePoint(Point pt, Point newOrigin)
        {
            return new Point(newOrigin.X + pt.X, newOrigin.Y - pt.Y);
        }

        private double TranslateAngle(double angle)
        {
            return -(angle - 90);
        }

        private Point GetMarkerPosition(double angle, double radius)
        {
            // Returns the coordinates of the point on the circumference
            // given an angle and the radius
            double angleRad = Math.PI / _flatCornerDegrees * angle;
            return new Point(radius * Math.Cos(angleRad), radius * Math.Sin(angleRad));
        }

        private void InitializeComponent()
        {
            AvaloniaXamlLoader.Load(this);

            // Create shapes, which will be repositioned in DrawClock
            _canvas = this.FindControl<Canvas>("ClockContainer");

            _clockBorder = new();
            _clockBorder.Fill = Brushes.Transparent;
            _clockBorder.Stroke = Brushes.Black;
            _clockBorder.StrokeThickness = _clockBorderTickness;
            _canvas.Children.Add(_clockBorder);

            // Create clock markers,
            // Hour markers are thicker than minute markers
            _clockMarkers = new Line[_clockMinutes];
            for (int currentMinute = 0; currentMinute < _clockMinutes; currentMinute++)
            {
                Line newMarker = new();
                newMarker.Fill = Brushes.Black;
                newMarker.Stroke = Brushes.Black;

                if (currentMinute % 5 == 0)
                    newMarker.StrokeThickness = _hourTickness;      // Hour (a thicker line "every 5 minutes")
                else
                    newMarker.StrokeThickness = _minuteTickness;    // Minute

                _clockMarkers[currentMinute] = newMarker;
                _canvas.Children.Add(newMarker);
            }

            // Create clock hands
            _handHours = new();
            _handHours.Fill = Brushes.Black;
            _handHours.Stroke = Brushes.Black;
            _handHours.StrokeThickness = _handHoursTickness;
            _canvas.Children.Add(_handHours);

            _handMinutes = new();
            _handMinutes.Fill = Brushes.Black;
            _handMinutes.Stroke = Brushes.Black;
            _handMinutes.StrokeThickness = _handMinutesTickness;
            _canvas.Children.Add(_handMinutes);

            _handSeconds = new();
            _handSeconds.Fill = Brushes.Red;
            _handSeconds.Stroke = Brushes.Red;
            _handSeconds.StrokeThickness = _handSecondsTickness;
            _canvas.Children.Add(_handSeconds);

            // Get notified when window size changes, so we can redraw the clock
            IObservable<double> obsWidth = this.GetObservable(WidthProperty);
            obsWidth.Subscribe((x) => RedrawClock());

            IObservable<double> obsHeight = this.GetObservable(HeightProperty);
            obsHeight.Subscribe((x) => RedrawClock());

            IObservable<WindowState> obsWindowState = this.GetObservable(WindowStateProperty);
            obsWindowState.Subscribe((x) => RedrawClock());

            // Start the timer
            _timer = new(new TimeSpan(0, 0, _clockFreqUpdatesSeconds), DispatcherPriority.Layout, (s, e) => RedrawClock());
            _timer.Start();
        }

        private void RedrawClock()
        {
            double clockRadius = GetClockRadius();
            double clockSize = GetClockSize();

            // Reposition and resize previously created shapes
            
            // Clock border
            _clockBorder.Width = clockSize;
            _clockBorder.Height = clockSize;

            Point pointClockPos = GetClockCenteredPosition();

            Canvas.SetLeft(_clockBorder, pointClockPos.X);
            Canvas.SetTop(_clockBorder, pointClockPos.Y);

            // Reposition clock markers, use the center of the window as new axis origin.
            // We assume that the y axis grows upward (see TranslatePoint).
            Point newOrigin = GetNewOrigin();

            for (int i = 0; i < _clockMinutes; i++)
            {
                Line marker = _clockMarkers[i];

                double hourMarkerLength = clockRadius - 25;
                double minuteMarkerLength = clockRadius - 10;

                Point startPos = TranslatePoint(GetMarkerPosition(i * _degreesPerMinute, clockRadius), newOrigin);

                Point endPos;
                if (i % 5 == 0)
                    endPos = TranslatePoint(GetMarkerPosition(i * _degreesPerMinute, hourMarkerLength), newOrigin);
                else
                    endPos = TranslatePoint(GetMarkerPosition(i * _degreesPerMinute, minuteMarkerLength), newOrigin);

                marker.StartPoint = startPos;
                marker.EndPoint = endPos;
            }

            RedrawHands(DateTime.Now);
        }

        private void RedrawHands(DateTime now)
        {
            // Here we redraw clock hands given a datetime
            // We use the GetMarkerPosition to calculate hands points.

            // Note the use of the TranslateAngle to adapt our reference system (which has the zero angle in correspondence of the y axis and rotates in reverse)
            // to that of trigonometric functions

            double clockRadius = GetClockRadius();
            Point newOrigin = GetNewOrigin();

            double secondDegrees = now.Second * _degreesPerSecond;
            double handSecondsLength1 = clockRadius - 20;
            double handSecondsLength2 = 40;
            Point startPointSec = TranslatePoint(GetMarkerPosition(TranslateAngle(secondDegrees), handSecondsLength1), newOrigin);
            Point endPointSec = TranslatePoint(GetMarkerPosition(TranslateAngle(secondDegrees + _flatCornerDegrees), handSecondsLength2), newOrigin);
            _handSeconds.StartPoint = startPointSec;
            _handSeconds.EndPoint = endPointSec;

            double minuteDegrees = now.Minute * _degreesPerMinute;
            double handMinutesLength1 = clockRadius - 20;
            double handMinutesLength2 = 40;
            Point startPointMin = TranslatePoint(GetMarkerPosition(TranslateAngle(minuteDegrees + (_degreesPerMinute * now.Second / _secondsPerMinute)), handMinutesLength1), newOrigin);
            Point endPointMin = TranslatePoint(GetMarkerPosition(TranslateAngle(minuteDegrees + (_degreesPerMinute * now.Second / _secondsPerMinute) + _flatCornerDegrees), handMinutesLength2), newOrigin);
            _handMinutes.StartPoint = startPointMin;
            _handMinutes.EndPoint = endPointMin;

            double hourDegrees = now.Hour * _degreesPerHour;
            double handHoursLength1 = clockRadius - 50;
            double handHoursLength2 = 30;
            Point startPointHour = TranslatePoint(GetMarkerPosition(TranslateAngle(hourDegrees + (_degreesPerHour * now.Minute / _minutesPerHour)), handHoursLength1), newOrigin);
            Point endPointHour = TranslatePoint(GetMarkerPosition(TranslateAngle(hourDegrees + (_degreesPerHour * now.Minute / _minutesPerHour) + _flatCornerDegrees), handHoursLength2), newOrigin);
            _handHours.StartPoint = startPointHour;
            _handHours.EndPoint = endPointHour;
        }
    }
}
